'---------------------------------------------------------------------
'  Copyright (C) Microsoft Corporation.  All rights reserved.
' 
'THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
'KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
'IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
'PARTICULAR PURPOSE.
'---------------------------------------------------------------------

Option Explicit On
Option Strict On

Imports System.Collections
Imports System.Collections.Generic
Imports System.Collections.ObjectModel
Imports System.ComponentModel
Imports System.ComponentModel.Design
Imports System.Diagnostics
Imports System.Drawing
Imports System.Drawing.Drawing2D
Imports System.Drawing.Design
Imports System.Windows.Forms
Imports System.Windows.Forms.Design
Imports System.Windows.Forms.Design.Behavior

Imports DesignSurfaceExtenderLibrary.Extensions

Namespace ConvexHullDesignSurfaceExtension

    ''' <summary>
    ''' An <see cref="ExtensionGlyph"/> which paints the convex hull
    ''' around the controls on the design surface.
    ''' </summary>
    ''' <remarks>The <see cref="ConvexHullGlyph.ComputeConvexHull"/> implementation is 
    ''' adapted from Carlos Moreno's solution, first published in the
    ''' November 1998 and December 1998 issues of C/C++ Users Journal:
    ''' http://www.mochima.com/articles/cuj_geometry_article/cuj_geometry_article.html
    ''' </remarks>
    Public Class ConvexHullGlyph
        Inherits ExtensionGlyph

        Private convexHullRegion As Region

        Private Shared customColor As Color = Color.FromArgb(50, Color.Gray)
        Private Shared shadowBrush As SolidBrush = New SolidBrush(customColor)

        Const right_turn As Integer = -1

        Friend Sub New( _
        ByVal behaviorService As BehaviorService, _
        ByVal changeService As IComponentChangeService, _
        ByVal selectionService As ISelectionService, _
        ByVal referenceService As IReferenceService, _
        ByVal serviceProvider As IServiceProvider, _
        ByVal adorner As Adorner)

            MyBase.New( _
                Nothing, _
                behaviorService, _
                changeService, _
                selectionService, _
                referenceService, _
                serviceProvider, _
                New ConvexHullBehavior(), _
                adorner)

        End Sub

        Protected Overrides Sub ComponentChangedEventHandler(ByVal sender As Object, ByVal e As ComponentChangedEventArgs)
            If TypeOf e.Component Is Control Then
                Dim propertyName As String = e.Member.Name

                If Me.DependentProperties.Contains(propertyName) Then
                    Me.ComputeGeometry()

                    If Me.adorner.Enabled Then
                        Me.adorner.Invalidate()
                    End If
                End If
            End If
        End Sub

        ''' <summary>
        ''' Converts the given collection of points to a <see cref="Region"/>.
        ''' </summary>
        Private Shared Function ConvertPointsToRegion(ByVal points As List(Of Point)) As Region
            Dim r As Region = Nothing

            If points.Count > 0 Then
                Dim gp As GraphicsPath = New GraphicsPath()
                gp.AddLines(points.ToArray())

                r = New Region(gp)
            End If

            Return r
        End Function

        Public Overrides Sub Paint(ByVal pe As PaintEventArgs)
            If Me.convexHullRegion IsNot Nothing Then
                pe.Graphics.FillRegion( _
                    ConvexHullGlyph.shadowBrush, _
                    Me.convexHullRegion)
            End If
        End Sub

        ''' <summary>
        ''' Calculates the convex hull around the controls on the design surface.
        ''' </summary>
        Protected Overrides Sub ComputeGeometry()
            Dim points As List(Of Point) = New List(Of Point)

            ' N.B.: This is not reliable when called from within 
            ' an IComponentChangeService event handler. The
            ' returned array may not include a newly added control
            ' or it may still include a recently deleted control.
            Dim controls As Object() = MyBase.referenceService.GetReferences(GetType(Control))

            For i As Integer = 0 To controls.Length - 1
                Dim c As Control = CType(controls(i), Control)

                If Not (TypeOf c Is Form) Then
                    Dim p As Point = Me.behaviorService.ControlToAdornerWindow(c)

                    Dim upperLeft As Point = p
                    Dim upperRight As Point = New Point(p.X + c.Width, p.Y)
                    Dim lowerLeft As Point = New Point(p.X, p.Y + c.Height)
                    Dim lowerRight As Point = New Point(p.X + c.Width, p.Y + c.Height)

                    points.Add(upperLeft)
                    points.Add(upperRight)
                    points.Add(lowerLeft)
                    points.Add(lowerRight)
                End If
            Next

            Dim convexHullPoints As List(Of Point) = ComputeConvexHull(points)

            Me.convexHullRegion = ConvertPointsToRegion(convexHullPoints)
        End Sub

        ''' <summary>
        ''' Calculates the convex hull around the controls on the design surface.
        ''' </summary>
        ''' <param name="controlToExclude">A control which is not included in
        ''' the convex hull computation.</param>
        ''' <remarks>This method is a workaround for the <see cref="IReferenceService"/>
        ''' issue. When <see cref="DesignSurfaceExtension.RemoveGlyphsForControl"/> is 
        ''' called from within 
        ''' <see cref="DesignSurfaceExtension.ComponentRemovedEventHandler"/>, 
        ''' the <see cref="IReferenceService.GetReferences"/> still reports the removed glyph.
        ''' This overload lets <see cref="DesignSurfaceExtension.RemoveGlyphsForControl"/>  
        ''' specify that the recently removed control be excluded from the geometry 
        ''' computation.
        ''' </remarks>
        Friend Overloads Sub ComputeGeometry(ByVal controlToExclude As Control)
            Dim points As List(Of Point) = New List(Of Point)

            Dim controls As Object() = MyBase.referenceService.GetReferences(GetType(Control))

            Dim controlsExcludingGivenControl As ArrayList = New ArrayList(controls)

            If controlsExcludingGivenControl.Contains(controlToExclude) Then
                controlsExcludingGivenControl.Remove(controlToExclude)

                controls = controlsExcludingGivenControl.ToArray()
            End If

            For i As Integer = 0 To controls.Length - 1
                Dim c As Control = CType(controls(i), Control)

                If Not (TypeOf c Is Form) Then
                    Dim p As Point = Me.behaviorService.ControlToAdornerWindow(c)

                    Dim upperLeft As Point = p
                    Dim upperRight As Point = New Point(p.X + c.Width, p.Y)
                    Dim lowerLeft As Point = New Point(p.X, p.Y + c.Height)
                    Dim lowerRight As Point = New Point(p.X + c.Width, p.Y + c.Height)

                    points.Add(upperLeft)
                    points.Add(upperRight)
                    points.Add(lowerLeft)
                    points.Add(lowerRight)
                End If
            Next

            Dim convexHullPoints As List(Of Point) = ComputeConvexHull(points)

            Me.convexHullRegion = ConvertPointsToRegion(convexHullPoints)
        End Sub

        Public Overrides Function GetHitTest(ByVal p As Point) As Cursor
            Return Nothing
        End Function

        Protected Overrides Function PopulateDependentProperties() As ReadOnlyCollection(Of String)
            Dim properties As List(Of String) = New List(Of String)()

            properties.Add("Size")
            properties.Add("Height")
            properties.Add("Width")
            properties.Add("Location")

            Return New ReadOnlyCollection(Of String)(properties)
        End Function

#Region "Convex Hull Implementation"

        ''' <summary>
        ''' Calculates the 2D convex hull of the given collection of points.
        ''' </summary>
        ''' <param name="points">A list of points for which to find the convex hull. </param>
        ''' <returns>A collection of points representing the convex hull of <paramref name="points"/>.</returns>
        ''' <remarks>Adapted from
        ''' http://www.mochima.com/articles/cuj_geometry_article/cuj_geometry_article.html.
        ''' </remarks>
        Friend Shared Function ComputeConvexHull(ByVal points As List(Of Point)) As List(Of Point)
            Dim convexHullPoints As List(Of Point) = New List(Of Point)()

            If points.Count > 0 Then
                Dim p_min_y As Point = points(0)

                For i As Integer = 1 To points.Count - 1
                    If points(i).Y < p_min_y.Y Then
                        p_min_y = points(i)
                    End If
                Next

                convexHullPoints.Add(p_min_y)

                Dim p As Point = p_min_y
                Dim new_p As Point = Point.Empty

                Do
                    If p <> points(0) Then
                        new_p = points(0)
                    Else
                        new_p = points(1)
                    End If

                    For i As Integer = 0 To points.Count - 1

                        If turn(p, new_p, points(i)) = right_turn Then
                            new_p = points(i)
                        End If
                    Next

                    If new_p <> p_min_y Then
                        convexHullPoints.Add(new_p)
                    End If

                    p = new_p

                Loop While (new_p <> p_min_y)

            End If

            Return convexHullPoints
        End Function

        Private Shared Function turn(ByVal p1 As Point, ByVal p2 As Point, ByVal p3 As Point) As Integer

            Dim signed_area As Double = p1.X * (p2.Y - p3.Y) + _
                                        p2.X * (p3.Y - p1.Y) + _
                                        p3.X * (p1.Y - p2.Y)

            Dim orientation As Integer
            If signed_area > 0 Then
                orientation = 1
            ElseIf signed_area < 0 Then
                orientation = -1
            Else
                orientation = 0
            End If

            Return orientation
        End Function

#End Region

    End Class

End Namespace
